#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for the `solve_qp` function."""

import unittest
import warnings

import numpy as np
import scipy
from numpy import array, dot, ones, random
from numpy.linalg import norm
from scipy.sparse import csc_matrix

from qpsolvers import (
    NoSolverSelected,
    ProblemError,
    SolverNotFound,
    available_solvers,
    solve_qp,
    sparse_solvers,
)

from .problems import get_qpmad_demo_problem

# Raising a ValueError when the problem is unbounded below is desired but not
# achieved by some solvers. Here are the behaviors observed as of March 2022.
# Unit tests only cover solvers that raise successfully:
behavior_on_unbounded = {
    "raise": ["cvxopt", "kvxopt", "ecos", "quadprog", "scs"],
    "return_crazy_solution": ["qpoases"],
    "return_none": ["osqp"],
}


class TestSolveQP(unittest.TestCase):
    """Test fixture for a variety of quadratic programs.

    Solver-specific tests are implemented in static methods called
    ``get_test_{foo}`` that return the test function for a given solver. The
    corresponding test function ``test_{foo}_{solver}`` is then added to the
    fixture below the class definition.
    """

    def setUp(self):
        """Prepare test fixture."""
        warnings.simplefilter("ignore", category=DeprecationWarning)
        warnings.simplefilter("ignore", category=UserWarning)

    def get_dense_problem(self):
        """Get dense problem as a sextuple of values to unpack.

        Returns
        -------
        P : numpy.ndarray
            Symmetric cost matrix .
        q : numpy.ndarray
            Cost vector.
        G : numpy.ndarray
            Linear inequality matrix.
        h : numpy.ndarray
            Linear inequality vector.
        A : numpy.ndarray, scipy.sparse.csc_matrix or cvxopt.spmatrix
            Linear equality matrix.
        b : numpy.ndarray
            Linear equality vector.
        """
        M = array([[1.0, 2.0, 0.0], [-8.0, 3.0, 2.0], [0.0, 1.0, 1.0]])
        P = dot(M.T, M)  # this is a positive definite matrix
        q = dot(array([3.0, 2.0, 3.0]), M).reshape((3,))
        G = array([[1.0, 2.0, 1.0], [2.0, 0.0, 1.0], [-1.0, 2.0, -1.0]])
        h = array([3.0, 2.0, -2.0]).reshape((3,))
        A = array([1.0, 1.0, 1.0])
        b = array([1.0])
        return P, q, G, h, A, b

    def get_sparse_problem(self):
        """Get sparse problem as a quadruplet of values to unpack.

        Returns
        -------
        P : scipy.sparse.csc_matrix
            Symmetric cost matrix.
        q : numpy.ndarray
            Cost vector.
        G : scipy.sparse.csc_matrix
            Linear inequality matrix.
        h : numpy.ndarray
            Linear inequality vector.
        """
        n = 150
        M = scipy.sparse.lil_matrix(scipy.sparse.eye(n))
        for i in range(1, n - 1):
            M[i, i + 1] = -1
            M[i, i - 1] = 1
        P = csc_matrix(M.dot(M.transpose()))
        q = -ones((n,))
        G = csc_matrix(-scipy.sparse.eye(n))
        h = -2.0 * ones((n,))
        return P, q, G, h

    def test_no_solver_selected(self):
        """Check that NoSolverSelected is raised when applicable."""
        P, q, G, h, A, b = self.get_dense_problem()
        with self.assertRaises(NoSolverSelected):
            solve_qp(P, q, G, h, A, b, solver=None)

    def test_solver_not_found(self):
        """SolverNotFound is raised when the solver does not exist."""
        P, q, G, h, A, b = self.get_dense_problem()
        with self.assertRaises(SolverNotFound):
            solve_qp(P, q, G, h, A, b, solver="ideal")

    @staticmethod
    def get_test(solver: str):
        """Get test function for a given solver.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h, A, b = self.get_dense_problem()
            x = solve_qp(P, q, G, h, A, b, solver=solver)
            x_sp = solve_qp(P, q, G, h, A, b, solver=solver)
            self.assertIsNotNone(x, f"{solver=}")
            self.assertIsNotNone(x_sp, f"{solver=}")
            known_solution = array([0.30769231, -0.69230769, 1.38461538])
            sol_tolerance = (
                5e-4
                if solver in ["osqp", "qpalm", "scs"]
                else (
                    1e-4
                    if solver in ["ecos", "jaxopt_osqp"]
                    else 5e-6 if solver in ["proxqp", "qpax", "sip"] else 1e-8
                )
            )
            eq_tolerance = 1e-5 if solver in ["jaxopt_osqp", "sip"] else 1e-10
            ineq_tolerance = (
                2e-4
                if solver in ["jaxopt_osqp", "qpalm", "scs"]
                else 5e-6 if solver in ["proxqp", "qpax"] else 1e-10
            )
            self.assertLess(
                norm(x - known_solution), sol_tolerance, f"{solver=}"
            )
            self.assertLess(
                norm(x_sp - known_solution), sol_tolerance, f"{solver=}"
            )
            self.assertLess(max(dot(G, x) - h), ineq_tolerance, f"{solver=}")
            self.assertLess(max(dot(A, x) - b), eq_tolerance, f"{solver=}")
            self.assertLess(min(dot(A, x) - b), eq_tolerance, f"{solver=}")

        return test

    @staticmethod
    def get_test_all_shapes(solver: str):
        """Get test function for a given solver.

        This variant tries all possible shapes for matrix and vector
        parameters.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.

        Note
        ----
        This function uses DAQP to find groundtruth solutions.
        """

        def test(self):
            P, q, G, h, _, _ = self.get_dense_problem()
            A = array([[1.0, 0.0, 0.0], [0.0, 0.4, 0.5]])
            b = array([-0.5, -1.2])
            lb = array([-0.5, -2, -0.8])
            ub = array([+1.0, +1.0, +1.0])

            ineq_variants = ((None, None), (G, h), (G[0], array([h[0]])))
            eq_variants = ((None, None), (A, b), (A[0], array([b[0]])))
            box_variants = ((None, None), (lb, None), (None, ub), (lb, ub))
            cases = [
                {
                    "P": P,
                    "q": q,
                    "G": G_case,
                    "h": h_case,
                    "A": A_case,
                    "b": b_case,
                    "lb": lb_case,
                    "ub": ub_case,
                }
                for (G_case, h_case) in ineq_variants
                for (A_case, b_case) in eq_variants
                for (lb_case, ub_case) in box_variants
            ]

            for i, test_case in enumerate(cases):
                no_inequality = "G" not in test_case or test_case["G"] is None
                if no_inequality and solver in ["qpswift", "qpax"]:
                    # QPs without inequality constraints are not handled by
                    # qpSWIFT or qpax
                    continue
                no_equality = "A" not in test_case or test_case["A"] is None
                if no_equality and solver in ["qpax"]:
                    # QPs without equality constraints not handled by qpax
                    continue
                has_one_equality = (
                    "A" in test_case
                    and test_case["A"] is not None
                    and test_case["A"].ndim == 1
                )
                has_lower_box = (
                    "lb" in test_case and test_case["lb"] is not None
                )
                if has_one_equality and has_lower_box and solver == "hpipm":
                    # Skipping this test for HPIPM for now
                    # See https://github.com/giaf/hpipm/issues/136
                    continue
                test_comp = {
                    k: v.shape if v is not None else "None"
                    for k, v in test_case.items()
                }
                daqp_solution = solve_qp(solver="daqp", **test_case)
                self.assertIsNotNone(
                    daqp_solution,
                    f"Baseline failed on parameters: {test_comp}",
                )
                solver_solution = solve_qp(solver=solver, **test_case)
                sol_tolerance = (
                    2e-2
                    if solver == "proxqp"
                    else (
                        5e-3
                        if solver in ["jaxopt_osqp"]
                        else (
                            2e-3
                            if solver in ["osqp", "qpalm", "scs"]
                            else 5e-4 if solver == "ecos" else 2e-4
                        )
                    )
                )
                self.assertLess(
                    norm(solver_solution - daqp_solution),
                    sol_tolerance,
                    f"Solver failed on parameters: {test_comp}",
                )

        return test

    @staticmethod
    def get_test_bounds(solver: str):
        """Get test function for a given solver.

        This variant adds vector bounds.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h, A, b = self.get_dense_problem()
            lb = array([-1.0, -2.0, -0.5])
            ub = array([1.0, -0.2, 1.0])
            x = solve_qp(P, q, G, h, A, b, lb, ub, solver=solver)
            self.assertIsNotNone(x)
            known_solution = array([0.41463415, -0.41463415, 1.0])
            sol_tolerance = (
                2e-3
                if solver == "proxqp"
                else (
                    5e-3
                    if solver in ["jaxopt_osqp", "osqp"]
                    else (
                        5e-5
                        if solver == "scs"
                        else (
                            1e-6
                            if solver in ["qpalm", "ecos", "qpax", "sip"]
                            else 1e-8
                        )
                    )
                )
            )
            eq_tolerance = 1e-5 if solver in ["proxqp", "sip"] else 1e-10
            ineq_tolerance = 1e-5 if solver == "proxqp" else 1e-10
            self.assertLess(
                norm(x - known_solution), sol_tolerance, f"{solver=}"
            )
            self.assertLess(max(dot(G, x) - h), ineq_tolerance, f"{solver=}")
            self.assertLess(max(dot(A, x) - b), eq_tolerance, f"{solver=}")
            self.assertLess(min(dot(A, x) - b), eq_tolerance, f"{solver=}")

        return test

    @staticmethod
    def get_test_no_cons(solver: str):
        """Get test function for a given solver.

        In this variant, there is no equality nor inequality constraint.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h, A, b = self.get_dense_problem()
            x = solve_qp(P, q, solver=solver)
            self.assertIsNotNone(x)
            known_solution = array([-0.64705882, -1.17647059, -1.82352941])
            sol_tolerance = (
                1e-3
                if solver == "ecos"
                else 1e-5 if solver in ["osqp", "qpalm"] else 1e-6
            )
            self.assertLess(
                norm(x - known_solution), sol_tolerance, f"{solver=}"
            )

        return test

    @staticmethod
    def get_test_no_eq(solver: str):
        """Get test function for a given solver.

        In this variant, there is no equality constraint.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h, A, b = self.get_dense_problem()
            x = solve_qp(P, q, G, h, solver=solver)
            self.assertIsNotNone(x)
            known_solution = array([-0.49025721, -1.57755261, -0.66484801])
            sol_tolerance = (
                1e-3
                if solver in ["ecos", "jaxopt_osqp"]
                else (
                    1e-4
                    if solver in ["qpalm", "sip"]
                    else 1e-5 if solver == "osqp" else 1e-6
                )
            )
            ineq_tolerance = (
                1e-3
                if solver == "jaxopt_osqp"
                else (
                    1e-5
                    if solver == "osqp"
                    else (
                        1e-6
                        if solver == "proxqp"
                        else 1e-7 if solver == "scs" else 1e-10
                    )
                )
            )
            self.assertLess(
                norm(x - known_solution), sol_tolerance, f"{solver=}"
            )
            self.assertLess(max(dot(G, x) - h), ineq_tolerance, f"{solver=}")

        return test

    @staticmethod
    def get_test_no_ineq(solver: str):
        """Get test function for a given solver.

        In this variant, there is no inequality constraint.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h, A, b = self.get_dense_problem()
            x = solve_qp(P, q, A=A, b=b, solver=solver)
            self.assertIsNotNone(x)
            known_solution = array([0.28026906, -1.55156951, 2.27130045])
            sol_tolerance = (
                1e-3
                if solver in ["jaxopt_osqp", "osqp", "qpalm"]
                else (
                    1e-5
                    if solver in ["ecos", "scs", "qpax"]
                    else (
                        1e-6
                        if solver == "highs"
                        else 1e-7 if solver == "proxqp" else 1e-8
                    )
                )
            )
            eq_tolerance = (
                1e-5
                if solver == "jaxopt_osqp"
                else (
                    1e-6
                    if solver in ["qpax", "sip"]
                    else 1e-7 if solver == "osqp" else 1e-9
                )
            )
            self.assertLess(
                norm(x - known_solution), sol_tolerance, f"{solver=}"
            )
            self.assertLess(max(dot(A, x) - b), eq_tolerance, f"{solver=}")
            self.assertLess(min(dot(A, x) - b), eq_tolerance, f"{solver=}")

        return test

    @staticmethod
    def get_test_one_ineq(solver: str):
        """Get test function for a given solver.

        In this variant, there is only one inequality constraint.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h, A, b = self.get_dense_problem()
            G, h = G[1], h[1].reshape((1,))
            x = solve_qp(P, q, G, h, A, b, solver=solver)
            self.assertIsNotNone(x)
            known_solution = array([0.30769231, -0.69230769, 1.38461538])
            sol_tolerance = (
                1e-3
                if solver in ["jaxopt_osqp", "qpalm"]
                else (
                    5e-4
                    if solver == "osqp"
                    else (
                        1e-5
                        if solver == "scs"
                        else (
                            5e-6
                            if solver in ["proxqp", "sip"]
                            else (
                                1e-6
                                if solver
                                in ["cvxopt", "kvxopt", "ecos", "qpax"]
                                else 5e-8 if solver == "qpswift" else 1e-8
                            )
                        )
                    )
                )
            )
            eq_tolerance = (
                1e-3
                if solver in ["jaxopt_osqp"]
                else (
                    1e-4
                    if solver in ["qpalm", "qpax", "sip"]
                    else 1e-8 if solver in ["osqp", "scs"] else 1e-10
                )
            )
            ineq_tolerance = (
                1e-5
                if solver == "proxqp"
                else (
                    1e-6
                    if solver == "qpax"
                    else (1e-7 if solver in ["qpswift", "scs"] else 1e-8)
                )
            )
            self.assertLess(
                norm(x - known_solution), sol_tolerance, f"{solver=}"
            )
            self.assertLess(max(dot(G, x) - h), ineq_tolerance, f"{solver=}")
            self.assertLess(max(dot(A, x) - b), eq_tolerance, f"{solver=}")
            self.assertLess(min(dot(A, x) - b), eq_tolerance, f"{solver=}")

        return test

    @staticmethod
    def get_test_sparse(solver: str):
        """Get test function for a given solver.

        This variant tests a sparse problem.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h = self.get_sparse_problem()
            kwargs = {}
            tol_solvers = ("osqp", "proxqp", "qpalm", "scs")
            if solver in tol_solvers:
                kwargs["eps_abs"] = 2e-4
            x = solve_qp(P, q, G, h, solver=solver)
            self.assertIsNotNone(x)
            known_solution = array([2.0] * 149 + [3.0])
            sol_tolerance = (
                5e-3
                if solver == "cvxopt" or solver == "kvxopt"
                else (
                    2e-3
                    if solver in ["osqp", "qpalm", "qpax"]
                    else (
                        1e-3
                        if solver in ["gurobi", "piqp"]
                        else (
                            5e-4
                            if solver in ["clarabel", "mosek"]
                            else (
                                1e-4
                                if solver in ["scs", "sip"]
                                else (
                                    2e-5
                                    if solver == "proxqp"
                                    else 1e-6 if solver == "highs" else 1e-7
                                )
                            )
                        )
                    )
                )
            )
            ineq_tolerance = 1e-4 if solver in tol_solvers else 1e-7
            self.assertLess(
                norm(x - known_solution), sol_tolerance, f"{solver=}"
            )
            self.assertLess(max(G * x - h), ineq_tolerance, f"{solver=}")

        return test

    @staticmethod
    def get_test_sparse_bounds(solver: str):
        """Get test function for a given solver.

        This variant tests a sparse problem with additional vector lower and
        upper bounds.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h = self.get_sparse_problem()
            lb = +2.2 * ones(q.shape)
            ub = +2.4 * ones(q.shape)
            x = solve_qp(P, q, G, h, lb=lb, ub=ub, solver=solver)
            self.assertIsNotNone(x)
            known_solution = array([2.2] * 149 + [2.4])
            sol_tolerance = (
                1e-3
                if solver in ["gurobi", "osqp", "qpalm", "sip"]
                else (
                    5e-6
                    if solver in ["mosek", "proxqp"]
                    else (
                        1e-7 if solver in ["cvxopt", "kvxopt", "scs"] else 1e-8
                    )
                )
            )
            ineq_tolerance = 1e-10
            self.assertLess(
                norm(x - known_solution), sol_tolerance, f"{solver=}"
            )
            self.assertLess(max(G * x - h), ineq_tolerance, f"{solver=}")

        return test

    @staticmethod
    def get_test_sparse_unfeasible(solver: str):
        """Get test function for a given solver.

        This variant tests an unfeasible sparse problem with additional vector
        lower and upper bounds.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h = self.get_sparse_problem()
            lb = +0.5 * ones(q.shape)
            ub = +1.5 * ones(q.shape)
            if solver == "cvxopt" or solver == "kvxopt":
                # Skipping this test for CVXOPT and KVXOPT for now
                # See https://github.com/cvxopt/cvxopt/issues/229
                return
            x = solve_qp(P, q, G, h, lb=lb, ub=ub, solver=solver)
            self.assertIsNone(x)

        return test

    @staticmethod
    def get_test_warmstart(solver: str):
        """Get test function for a given solver.

        This variant warm starts.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h, A, b = self.get_dense_problem()
            known_solution = array([0.30769231, -0.69230769, 1.38461538])
            initvals = known_solution + 0.1 * random.random(3)
            x = solve_qp(
                P,
                q,
                G,
                h,
                A,
                b,
                solver=solver,
                initvals=initvals,
                verbose=True,  # increases coverage
            )
            self.assertIsNotNone(x)
            sol_tolerance = (
                1e-3
                if solver in ["osqp", "qpalm", "scs"]
                else (
                    1e-4
                    if solver in ["ecos", "jaxopt_osqp"]
                    else 5e-6 if solver in ["proxqp", "qpax", "sip"] else 1e-8
                )
            )
            eq_tolerance = (
                1e-4 if solver in ["jaxopt_osqp", "osqp", "sip"] else 1e-10
            )
            ineq_tolerance = (
                1e-3
                if solver in ["osqp", "qpalm", "scs"]
                else (
                    1e-5
                    if solver in ["jaxopt_osqp", "proxqp", "qpax"]
                    else 1e-10
                )
            )
            self.assertLess(
                norm(x - known_solution), sol_tolerance, f"{solver=}"
            )
            self.assertLess(max(dot(G, x) - h), ineq_tolerance, f"{solver=}")
            self.assertLess(max(dot(A, x) - b), eq_tolerance, f"{solver=}")
            self.assertLess(min(dot(A, x) - b), eq_tolerance, f"{solver=}")

        return test

    @staticmethod
    def get_test_raise_on_unbounded_below(solver: str):
        """ValueError is raised when the problem is unbounded below.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.

        Notes
        -----
        Detecting non-convexity is not a trivial problem and most solvers leave
        it to the user. See for instance the `recommendation from OSQP
        <https://osqp.org/docs/interfaces/status_values.html#status-values>`_.
        We only run this test for functions that successfully detect unbounded
        problems when the eigenvalues of :math:`P` are close to zero.
        """

        def test(self):
            v = array([5.4, -1.2, -1e-2, 1e4])
            P = dot(v.reshape(4, 1), v.reshape(1, 4))
            q = array([-1.0, -2, 0, 3e-4])
            # q is in the nullspace of P, so the problem is unbounded below
            with self.assertRaises(ProblemError):
                solve_qp(P, q, solver=solver)

        return test

    @staticmethod
    def get_test_qpmad_demo(solver: str):
        """Get test function for a given solver.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            problem = get_qpmad_demo_problem()
            P, q, G, h, _, _, lb, ub = problem.unpack()
            x = solve_qp(P, q, G, h, lb=lb, ub=ub, solver=solver)
            known_solution = array(
                [
                    1.0,
                    2.0,
                    3.0,
                    4.0,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                    -0.71875,
                ]
            )
            sol_tolerance = (
                1e-2
                if solver in ["jaxopt_osqp", "osqp"]
                else (
                    5e-4
                    if solver in ["qpalm", "scs", "qpax"]
                    else (
                        2e-5
                        if solver == "proxqp"
                        else (
                            1e-6
                            if solver
                            in [
                                "cvxopt",
                                "kvxopt",
                                "mosek",
                                "qpswift",
                                "piqp",
                                "sip",
                            ]
                            else 1e-8
                        )
                    )
                )
            )
            self.assertIsNotNone(x)
            self.assertLess(np.linalg.norm(x - known_solution), sol_tolerance)

        return test


# Generate test fixtures for each solver
for solver in available_solvers:
    setattr(TestSolveQP, f"test_{solver}", TestSolveQP.get_test(solver))
    setattr(
        TestSolveQP,
        f"test_all_shapes_{solver}",
        TestSolveQP.get_test_all_shapes(solver),
    )
    setattr(
        TestSolveQP,
        f"test_bounds_{solver}",
        TestSolveQP.get_test_bounds(solver),
    )
    if solver not in ["qpswift"]:
        # qpSWIFT: https://github.com/qpSWIFT/qpSWIFT/issues/2
        setattr(
            TestSolveQP,
            f"test_no_cons_{solver}",
            TestSolveQP.get_test_no_cons(solver),
        )
    setattr(
        TestSolveQP,
        f"test_no_eq_{solver}",
        TestSolveQP.get_test_no_eq(solver),
    )
    if solver not in ["qpswift"]:
        # qpSWIFT: https://github.com/qpSWIFT/qpSWIFT/issues/2
        setattr(
            TestSolveQP,
            f"test_no_ineq_{solver}",
            TestSolveQP.get_test_no_ineq(solver),
        )
    setattr(
        TestSolveQP,
        f"test_one_ineq_{solver}",
        TestSolveQP.get_test_one_ineq(solver),
    )
    if solver in sparse_solvers:
        setattr(
            TestSolveQP,
            f"test_sparse_{solver}",
            TestSolveQP.get_test_sparse(solver),
        )
        setattr(
            TestSolveQP,
            f"test_sparse_bounds_{solver}",
            TestSolveQP.get_test_sparse_bounds(solver),
        )
        setattr(
            TestSolveQP,
            f"test_sparse_unfeasible_{solver}",
            TestSolveQP.get_test_sparse_unfeasible(solver),
        )
    setattr(
        TestSolveQP,
        f"test_warmstart_{solver}",
        TestSolveQP.get_test_warmstart(solver),
    )
    if solver in behavior_on_unbounded["raise"]:
        setattr(
            TestSolveQP,
            f"test_raise_on_unbounded_below_{solver}",
            TestSolveQP.get_test_raise_on_unbounded_below(solver),
        )
    setattr(
        TestSolveQP,
        f"test_qpmad_demo_{solver}",
        TestSolveQP.get_test_qpmad_demo(solver),
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for ECOS."""

import unittest
import warnings

import numpy as np

from qpsolvers.exceptions import ProblemError

from .problems import get_sd3310_problem

try:
    from qpsolvers.solvers.ecos_ import ecos_solve_qp

    class TestECOS(unittest.TestCase):
        """Tests specific to ECOS."""

        def test_problem(self):
            problem = get_sd3310_problem()
            P, q, G, h, A, b, lb, ub = problem.unpack()
            self.assertIsNotNone(ecos_solve_qp(P, q, G, h, A, b, lb, ub))

        def test_infinite_inequality(self):
            problem = get_sd3310_problem()
            P, q, G, h, A, b, _, _ = problem.unpack()
            lb = np.array([-1.0, -np.inf, -1.0])
            ub = np.array([np.inf, 1.0, 1.0])
            with self.assertRaises(ProblemError):
                ecos_solve_qp(P, q, G, h, lb=lb, ub=ub)

except ImportError as exn:  # solver not installed
    warnings.warn(f"Skipping ECOS tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Convert problems from and to standard QP form."""

from .combine_linear_box_inequalities import combine_linear_box_inequalities
from .ensure_sparse_matrices import ensure_sparse_matrices
from .linear_from_box_inequalities import linear_from_box_inequalities
from .socp_from_qp import socp_from_qp
from .split_dual_linear_box import split_dual_linear_box

__all__ = [
    "combine_linear_box_inequalities",
    "linear_from_box_inequalities",
    "socp_from_qp",
    "split_dual_linear_box",
    "ensure_sparse_matrices",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for Gurobi."""

import unittest
import warnings

from .problems import get_sd3310_problem

try:
    from qpsolvers.solvers.gurobi_ import gurobi_solve_qp

    class TestGurobi(unittest.TestCase):
        """Test fixture for the Gurobi solver."""

        def test_gurobi_params(self):
            problem = get_sd3310_problem()
            x = gurobi_solve_qp(
                problem.P,
                problem.q,
                problem.G,
                problem.h,
                TimeLimit=0.1,
                FeasibilityTol=1e-8,
            )
            self.assertIsNotNone(x)

except ImportError as exn:  # solver not installed
    warnings.warn(f"Skipping Gurobi tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

import numpy as np

from qpsolvers import Problem


def get_sd3310_problem() -> Problem:
    """
    Get a small dense problem with 3 optimization variables, 3 inequality
    constraints, 1 equality constraint and 0 box constraint.
    """
    M = np.array([[1.0, 2.0, 0.0], [-8.0, 3.0, 2.0], [0.0, 1.0, 1.0]])
    P = np.dot(M.T, M)  # this is a positive definite matrix
    q = np.dot(np.array([3.0, 2.0, 3.0]), M).reshape((3,))
    G = np.array([[1.0, 2.0, 1.0], [2.0, 0.0, 1.0], [-1.0, 2.0, -1.0]])
    h = np.array([3.0, 2.0, -2.0]).reshape((3,))
    A = np.array([1.0, 1.0, 1.0])
    b = np.array([1.0])
    return Problem(P, q, G, h, A, b)


def get_qpmad_demo_problem():
    """
    Problem from qpmad's `demo.cpp
    <https://github.com/asherikov/qpmad/blob/5e4038f15d85a2a396bb062599f9d7a06d0b0764/test/dependency/demo.cpp>`__.
    """
    P = np.eye(20)
    q = np.ones((20,))
    G = np.vstack([np.ones((1, 20)), -np.ones((1, 20))])
    h = np.hstack([1.5, 1.5])
    lb = np.array(
        [
            1.0,
            2.0,
            3.0,
            4.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
            -5.0,
        ]
    )
    ub = np.array(
        [
            1.0,
            2.0,
            3.0,
            4.0,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
        ]
    )
    return Problem(P, q, G, h, lb=lb, ub=ub)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for the `solve_qp` function."""

import unittest
import warnings

import numpy as np

from qpsolvers import available_solvers
from qpsolvers.conversions import combine_linear_box_inequalities


class TestCombineLinearBoxInequalities(unittest.TestCase):
    def setUp(self):
        """Prepare test fixture."""
        warnings.simplefilter("ignore", category=DeprecationWarning)
        warnings.simplefilter("ignore", category=UserWarning)

    def get_dense_problem(self):
        """Get dense problem as a sextuple of values to unpack.

        Returns
        -------
        P : numpy.ndarray
            Symmetric cost matrix .
        q : numpy.ndarray
            Cost vector.
        G : numpy.ndarray
            Linear inequality matrix.
        h : numpy.ndarray
            Linear inequality vector.
        A : numpy.ndarray, scipy.sparse.csc_matrix or cvxopt.spmatrix
            Linear equality matrix.
        b : numpy.ndarray
            Linear equality vector.
        """
        M = np.array([[1.0, 2.0, 0.0], [-8.0, 3.0, 2.0], [0.0, 1.0, 1.0]])
        P = np.dot(M.T, M)  # this is a positive definite matrix
        q = np.dot(np.array([3.0, 2.0, 3.0]), M).reshape((3,))
        G = np.array([[1.0, 2.0, 1.0], [2.0, 0.0, 1.0], [-1.0, 2.0, -1.0]])
        h = np.array([3.0, 2.0, -2.0]).reshape((3,))
        A = np.array([1.0, 1.0, 1.0])
        b = np.array([1.0])
        return P, q, G, h, A, b

    @staticmethod
    def get_test_all_shapes(solver: str):
        """Get test function for a given solver.

        This variant tries all possible shapes for matrix and vector
        parameters.

        Parameters
        ----------
        solver :
            Name of the solver to test.

        Returns
        -------
        :
            Test function for that solver.
        """

        def test(self):
            P, q, G, h, _, _ = self.get_dense_problem()
            A = np.array([[1.0, 0.0, 0.0], [0.0, 0.4, 0.5]])
            b = np.array([-0.5, -1.2])
            lb = np.array([-0.5, -2, -0.8])
            ub = np.array([+1.0, +1.0, +1.0])

            ineq_variants = ((None, None), (G, h), (G[0], np.array([h[0]])))
            eq_variants = ((None, None), (A, b), (A[0], np.array([b[0]])))
            box_variants = ((None, None), (lb, None), (None, ub), (lb, ub))
            cases = [
                {
                    "P": P,
                    "q": q,
                    "G": G_case,
                    "h": h_case,
                    "A": A_case,
                    "b": b_case,
                    "lb": lb_case,
                    "ub": ub_case,
                }
                for (G_case, h_case) in ineq_variants
                for (A_case, b_case) in eq_variants
                for (lb_case, ub_case) in box_variants
            ]

            for i, test_case in enumerate(cases):
                G = test_case["G"]
                h = test_case["h"]
                lb = test_case["lb"]
                ub = test_case["ub"]
                n = test_case["q"].shape[0]
                if G is None and lb is None and ub is None:
                    continue
                elif isinstance(G, np.ndarray) and G.ndim == 1:
                    G = G.reshape((1, G.shape[0]))
                C, u, l = combine_linear_box_inequalities(
                    G, h, lb, ub, n, use_csc=False
                )
                self.assertTrue(isinstance(C, np.ndarray))
                self.assertTrue(isinstance(u, np.ndarray))
                self.assertTrue(isinstance(l, np.ndarray))

        return test


# Generate test fixtures for each solver
for solver in available_solvers:
    setattr(
        TestCombineLinearBoxInequalities,
        f"test_all_shapes_{solver}",
        TestCombineLinearBoxInequalities.get_test_all_shapes(solver),
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2023 Stéphane Caron and the qpsolvers contributors

"""Import unsupported QP solvers."""

from typing import Any, Callable, Dict, List, Optional

from numpy import ndarray

from ..problem import Problem
from ..solution import Solution

available_solvers: List[str] = []
dense_solvers: List[str] = []
solve_function: Dict[str, Any] = {}
sparse_solvers: List[str] = []


# NPPro
# =====

nppro_solve_problem: Optional[
    Callable[
        [
            Problem,
            Optional[ndarray],
        ],
        Solution,
    ]
] = None

nppro_solve_qp: Optional[
    Callable[
        [
            ndarray,
            ndarray,
            Optional[ndarray],
            Optional[ndarray],
            Optional[ndarray],
            Optional[ndarray],
            Optional[ndarray],
            Optional[ndarray],
            Optional[ndarray],
        ],
        Optional[ndarray],
    ]
] = None

try:
    from .nppro_ import nppro_solve_problem, nppro_solve_qp

    solve_function["nppro"] = nppro_solve_problem
    available_solvers.append("nppro")
    dense_solvers.append("nppro")
except ImportError:
    pass


__all__ = [
    "available_solvers",
    "nppro_solve_qp",
    "solve_function",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Solver interface for `HiGHS <https://github.com/ERGO-Code/HiGHS>`__.

HiGHS is an open source, serial and parallel solver for large scale sparse
linear programming (LP), mixed-integer programming (MIP), and quadratic
programming (QP). It is written mostly in C++11 and available under the MIT
licence. HiGHS's QP solver implements a Nullspace Active Set method. It works
best on moderately-sized dense problems. If you are using HiGHS in a scientific
work, consider citing the corresponding paper [Huangfu2018]_.
"""

import warnings
from typing import Optional, Union

import highspy
import numpy as np
import scipy.sparse as spa

from ..conversions import ensure_sparse_matrices
from ..problem import Problem
from ..solution import Solution


def __set_hessian(model: highspy.HighsModel, P: spa.csc_matrix) -> None:
    """Set Hessian :math:`Q` of the cost :math:`(1/2) x^T Q x + c^T x`.

    Parameters
    ----------
    model :
        HiGHS model.
    P :
        Positive semidefinite cost matrix.
    """
    model.hessian_.dim_ = P.shape[0]
    model.hessian_.start_ = P.indptr
    model.hessian_.index_ = P.indices
    model.hessian_.value_ = P.data


def __set_columns(
    model: highspy.HighsModel,
    q: np.ndarray,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
) -> None:
    r"""Set columns of the model.

    Columns consist of:

    - Linear part :math:`c` of the cost :math:`(1/2) x^T Q x + c^T x`
    - Box inequalities :math:`l \leq x \leq u``

    Parameters
    ----------
    model :
        HiGHS model.
    q :
        Cost vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    """
    n = q.shape[0]
    lp = model.lp_
    lp.num_col_ = n
    lp.col_cost_ = q
    lp.col_lower_ = lb if lb is not None else np.full((n,), -highspy.kHighsInf)
    lp.col_upper_ = ub if ub is not None else np.full((n,), highspy.kHighsInf)


def __set_rows(
    model: highspy.HighsModel,
    G: Optional[spa.csc_matrix] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[spa.csc_matrix] = None,
    b: Optional[np.ndarray] = None,
) -> None:
    r"""Set rows :math:`L \leq A x \leq U`` of the model.

    Parameters
    ----------
    model :
        HiGHS model.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    """
    lp = model.lp_
    lp.num_row_ = 0
    row_list: list = []
    row_lower: list = []
    row_upper: list = []
    if G is not None:
        lp.num_row_ += G.shape[0]
        row_list.append(G)
        row_lower.append(np.full((G.shape[0],), -highspy.kHighsInf))
        row_upper.append(h)
    if A is not None:
        lp.num_row_ += A.shape[0]
        row_list.append(A)
        row_lower.append(b)
        row_upper.append(b)
    if not row_list:
        return
    row_matrix = spa.vstack(row_list, format="csc")
    lp.a_matrix_.format_ = highspy.MatrixFormat.kColwise
    lp.a_matrix_.start_ = row_matrix.indptr
    lp.a_matrix_.index_ = row_matrix.indices
    lp.a_matrix_.value_ = row_matrix.data
    lp.a_matrix_.num_row_ = row_matrix.shape[0]
    lp.a_matrix_.num_col_ = row_matrix.shape[1]
    lp.row_lower_ = np.hstack(row_lower)
    lp.row_upper_ = np.hstack(row_upper)


def highs_solve_problem(
    problem: Problem,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Solution:
    """Solve a quadratic program using HiGHS.

    Parameters
    ----------
    problem :
        Quadratic program to solve.
    initvals :
        Warm-start guess vector for the primal solution.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution returned by the solver.

    Notes
    -----
    Keyword arguments are forwarded to HiGHS as options. For instance, we
    can call ``highs_solve_qp(P, q, G, h, u, primal_feasibility_tolerance=1e-8,
    dual_feasibility_tolerance=1e-8)``. HiGHS settings include the following:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Description
       * - ``dual_feasibility_tolerance``
         - Dual feasibility tolerance.
       * - ``primal_feasibility_tolerance``
         - Primal feasibility tolerance.
       * - ``time_limit``
         - Run time limit in seconds.

    Check out the `HiGHS documentation <https://ergo-code.github.io/HiGHS/>`_
    for more information on the solver.
    """
    P, q, G, h, A, b, lb, ub = problem.unpack()
    P, G, A = ensure_sparse_matrices("highs", P, G, A)
    if initvals is not None:
        warnings.warn(
            "HiGHS: warm-start values are not available for this solver, "
            "see: https://github.com/qpsolvers/qpsolvers/issues/94"
        )

    model = highspy.HighsModel()
    __set_hessian(model, P)
    __set_columns(model, q, lb, ub)
    __set_rows(model, G, h, A, b)

    solver = highspy.Highs()
    if verbose:
        solver.setOptionValue("log_to_console", True)
        solver.setOptionValue("log_dev_level", highspy.HighsLogType.kVerbose)
        solver.setOptionValue(
            "highs_debug_level", highspy.HighsLogType.kVerbose
        )
    else:  # not verbose
        solver.setOptionValue("log_to_console", False)
    for option, value in kwargs.items():
        solver.setOptionValue(option, value)
    solver.passModel(model)
    solver.run()

    result = solver.getSolution()
    model_status = solver.getModelStatus()

    solution = Solution(problem)
    solution.found = model_status == highspy.HighsModelStatus.kOptimal
    solution.x = np.array(result.col_value)
    if G is not None:
        solution.z = -np.array(result.row_dual[: G.shape[0]])
        solution.y = (
            -np.array(result.row_dual[G.shape[0] :])
            if A is not None
            else np.empty((0,))
        )
    else:  # G is None
        solution.z = np.empty((0,))
        solution.y = (
            -np.array(result.row_dual) if A is not None else np.empty((0,))
        )
    solution.z_box = (
        -np.array(result.col_dual)
        if lb is not None or ub is not None
        else np.empty((0,))
    )
    return solution


def highs_solve_qp(
    P: Union[np.ndarray, spa.csc_matrix],
    q: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program using HiGHS.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{x}{\mbox{minimize}} &
                \frac{1}{2} x^T P x + q^T x \\
            \mbox{subject to}
                & G x \leq h                \\
                & A x = b                   \\
                & lb \leq x \leq ub
        \end{array}\end{split}

    It is solved using `HiGHS <https://github.com/ERGO-Code/HiGHS>`__.

    Parameters
    ----------
    P :
        Positive semidefinite cost matrix.
    q :
        Cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    initvals :
        Warm-start guess vector for the primal solution.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP, if found, otherwise ``None``.

    Notes
    -----
    Keyword arguments are forwarded to HiGHS as options. For instance, we
    can call ``highs_solve_qp(P, q, G, h, u, primal_feasibility_tolerance=1e-8,
    dual_feasibility_tolerance=1e-8)``. HiGHS settings include the following:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Description
       * - ``dual_feasibility_tolerance``
         - Dual feasibility tolerance.
       * - ``primal_feasibility_tolerance``
         - Primal feasibility tolerance.
       * - ``time_limit``
         - Run time limit in seconds.

    Check out the `HiGHS documentation <https://ergo-code.github.io/HiGHS/>`_
    for more information on the solver.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = highs_solve_problem(problem, initvals, verbose, **kwargs)
    return solution.x if solution.found else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for quadprog."""

import unittest
import warnings

import numpy as np
import scipy.sparse as spa

from qpsolvers.exceptions import ProblemError

from .problems import get_sd3310_problem

try:
    from qpsolvers.solvers.quadprog_ import quadprog_solve_qp

    class TestQuadprog(unittest.TestCase):
        """Test fixture for the quadprog solver."""

        def setUp(self):
            """Prepare test fixture."""
            warnings.simplefilter("ignore", category=UserWarning)

        def test_non_psd_cost(self):
            problem = get_sd3310_problem()
            P, q, G, h, A, b, _, _ = problem.unpack()
            P -= np.eye(3)
            with self.assertRaises(ProblemError):
                quadprog_solve_qp(P, q, G, h, A, b)

        def test_quadprog_value_error(self):
            problem = get_sd3310_problem()
            P, q, G, h, A, b, _, _ = problem.unpack()
            q = q[1:]  # raise "G and a must have the same dimension"
            self.assertIsNone(quadprog_solve_qp(P, q, G, h, A, b))

        def test_not_sparse(self):
            """Raise a ProblemError on sparse problems."""
            problem = get_sd3310_problem()
            P, q, G, h, A, b, _, _ = problem.unpack()
            P = spa.csc_matrix(P)
            with self.assertRaises(ProblemError):
                quadprog_solve_qp(P, q, G, h, A, b)

except ImportError as exn:  # solver not installed
    warnings.warn(f"Skipping quadprog tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Test all available QP solvers on a model predictive control problem."""

from os.path import basename

from IPython import get_ipython
from qpsolvers import dense_solvers, sparse_solvers

from model_predictive_control import (
    HumanoidModelPredictiveControl,
    HumanoidSteppingProblem,
)

problem = HumanoidSteppingProblem()
mpc = HumanoidModelPredictiveControl(problem)


if __name__ == "__main__":
    if get_ipython() is None:
        print(
            "This example should be run with IPython:\n\n"
            f"\tipython -i {basename(__file__)}\n"
        )
        exit()

    dense_instr = {
        solver: f"u = mpc.solve(solver='{solver}', sparse=False)"
        for solver in dense_solvers
    }
    sparse_instr = {
        solver: f"u = mpc.solve(solver='{solver}', sparse=True)"
        for solver in sparse_solvers
    }

    benchmark = "https://github.com/qpsolvers/qpbenchmark"
    print("\nTesting QP solvers on one given model predictive control problem")
    print(f"For a proper benchmark, check out {benchmark}")

    print("\nDense solvers\n-------------")
    for solver, instr in dense_instr.items():
        print(f"{solver}: ", end="")
        get_ipython().run_line_magic("timeit", instr)

    print("\nSparse solvers\n--------------")
    for solver, instr in sparse_instr.items():
        print(f"{solver}: ", end="")
        get_ipython().run_line_magic("timeit", instr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2023 Inria

"""Active set: indices of inequality constraints saturated at the optimum."""

from dataclasses import dataclass
from typing import Optional, Sequence


@dataclass
class ActiveSet:
    """Indices of active inequality constraints.

    Attributes
    ----------
    G_indices :
        Indices of active linear inequality constraints.
    lb_indices :
        Indices of active lower-bound inequality constraints.
    ub_indices :
        Indices of active upper-bound inequality constraints.
    """

    G_indices: Sequence[int]
    lb_indices: Sequence[int]
    ub_indices: Sequence[int]

    def __init__(
        self,
        G_indices: Optional[Sequence[int]] = None,
        lb_indices: Optional[Sequence[int]] = None,
        ub_indices: Optional[Sequence[int]] = None,
    ) -> None:
        self.G_indices = list(G_indices) if G_indices is not None else []
        self.lb_indices = list(lb_indices) if lb_indices is not None else []
        self.ub_indices = list(ub_indices) if ub_indices is not None else []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Solver interface for `MOSEK <https://www.mosek.com/>`__.

MOSEK is a solver for linear, mixed-integer linear, quadratic, mixed-integer
quadratic, quadratically constraint, conic and convex nonlinear mathematical
optimization problems. Its interior-point method is geared towards large scale
sparse problems, in particular for linear or conic quadratic programs.
"""

import warnings
from typing import Optional, Union

import mosek
import numpy as np
import scipy.sparse as spa

from ..problem import Problem
from ..solution import Solution
from ..solve_unconstrained import solve_unconstrained
from ..solvers.cvxopt_ import cvxopt_solve_problem


def mosek_solve_problem(
    problem: Problem,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Solution:
    """Solve a quadratic program using MOSEK.

    Parameters
    ----------
    P :
        Symmetric cost matrix.
    q :
        Cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    initvals :
        Warm-start guess vector.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP, if found, otherwise ``None``.
    """
    if problem.is_unconstrained:
        warnings.warn(
            "QP is unconstrained: solving with SciPy's LSQR rather than MOSEK"
        )
        return solve_unconstrained(problem)
    if "mosek" not in kwargs:
        kwargs["mosek"] = {}
    kwargs["mosek"][mosek.iparam.log] = 1 if verbose else 0
    solution = cvxopt_solve_problem(problem, "mosek", initvals, **kwargs)
    return solution


def mosek_solve_qp(
    P: Union[np.ndarray, spa.csc_matrix],
    q: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program using MOSEK.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{x}{\mbox{minimize}} &
                \frac{1}{2} x^T P x + q^T x \\
            \mbox{subject to}
                & G x \leq h                \\
                & A x = b                   \\
                & lb \leq x \leq ub
        \end{array}\end{split}

    It is solved using the `MOSEK interface from CVXOPT
    <https://cvxopt.org/userguide/coneprog.html#optional-solvers>`_.

    Parameters
    ----------
    P :
        Symmetric cost matrix.
    q :
        Cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    initvals :
        Warm-start guess vector.
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP, if found, otherwise ``None``.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = mosek_solve_problem(problem, initvals, verbose, **kwargs)
    return solution.x if solution.found else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""
Exceptions from qpsolvers.

We catch all solver exceptions and re-throw them in a qpsolvers-owned exception
to avoid abstraction leakage. See this `design decision
<https://github.com/getparthenon/parthenon/wiki/Design-Decision:-Throw-Custom-Exceptions>`__
for more details on the rationale behind this choice.
"""


class QPError(Exception):
    """Base class for qpsolvers exceptions."""


class NoSolverSelected(QPError):
    """Exception raised when the `solver` keyword argument is not set."""


class ParamError(QPError):
    """Exception raised when solver parameters are incorrect."""


class ProblemError(QPError):
    """Exception raised when a quadratic program is malformed."""


class SolverNotFound(QPError):
    """Exception raised when a requested solver is not found."""


class SolverError(QPError):
    """Exception raised when a solver failed."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2024 Lev Kozlov

"""
Solver interface for jaxopt's implementation of the OSQP algorithm.

JAXopt is a library of hardware-accelerated, batchable and differentiable
optimizers implemented with JAX. JAX itself is a library for array-oriented
numerical computation that provides automatic differentiation and just-in-time
compilation.
"""

import warnings
from typing import Optional

import jax.numpy as jnp
import jaxopt
import numpy as np

from ..conversions import linear_from_box_inequalities, split_dual_linear_box
from ..problem import Problem
from ..solution import Solution
from ..solve_unconstrained import solve_unconstrained


def jaxopt_osqp_solve_problem(
    problem: Problem,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Solution:
    """Solve a quadratic program with the OSQP algorithm implemented in jaxopt.

    Parameters
    ----------
    problem :
        Quadratic program to solve.
    initvals :
        Warm-start guess vector (not used).
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP returned by the solver.

    Notes
    -----
    All other keyword arguments are forwarded as keyword arguments to
    jaxopt.OSQP. For instance, you can call ``jaxopt_osqp_solve_qp(P, q, G,
    h, sigma=1e-5, momentum=0.9)``.

    Note that JAX by default uses 32-bit floating point numbers, which can lead
    to numerical instability. If you encounter numerical issues, consider using
    64-bit floating point numbers by setting its `jax_enable_x64`
    configuration.
    """
    if problem.is_unconstrained:
        warnings.warn(
            "QP is unconstrained: "
            "solving with SciPy's LSQR rather than jaxopt's OSQP"
        )
        return solve_unconstrained(problem)

    P, q, G_0, h_0, A, b, lb, ub = problem.unpack()
    G, h = linear_from_box_inequalities(G_0, h_0, lb, ub, use_sparse=False)
    if initvals is not None and verbose:
        warnings.warn("warm-start values are ignored by this wrapper")

    osqp = jaxopt.OSQP(**kwargs)
    result = osqp.run(
        params_obj=(jnp.array(P), jnp.array(q)),
        params_eq=(jnp.array(A), jnp.array(b)) if A is not None else None,
        params_ineq=(jnp.array(G), jnp.array(h)) if G is not None else None,
    )

    solution = Solution(problem)
    solution.x = np.array(result.params.primal)
    solution.found = result.state.status == jaxopt.BoxOSQP.SOLVED
    solution.y = (
        np.array(result.params.dual_eq)
        if result.params.dual_eq is not None
        else np.empty((0,))
    )

    # split the dual variables into
    # the box constraints and the linear constraints
    solution.z, solution.z_box = split_dual_linear_box(
        np.array(result.params.dual_ineq), problem.lb, problem.ub
    )

    solution.extras = {
        "iter_num": int(result.state.iter_num),
        "error": float(result.state.error),
        "status": int(result.state.status),
    }

    return solution


def jaxopt_osqp_solve_qp(
    P: np.ndarray,
    q: np.ndarray,
    G: Optional[np.ndarray] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[np.ndarray] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a QP with the OSQP algorithm implemented in jaxopt.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
        \underset{\mbox{minimize}}{x} &
            \frac{1}{2} x^T P x + q^T x \\
        \mbox{subject to}
            & G x \leq h                \\
            & A x = b                   \\
            & lb \leq x \leq ub
        \end{array}\end{split}

    It is solved using `jaxopt.OSQP
    <https://jaxopt.github.io/stable/_autosummary/jaxopt.OSQP.html>`__.

    Parameters
    ----------
    P :
        Positive semidefinite cost matrix.
    q :
        Cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    verbose :
        Set to `True` to print out extra information.
    initvals :
        Warm-start guess vector. Not used.

    Returns
    -------
    :
        Primal solution to the QP, if found, otherwise ``None``.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = jaxopt_osqp_solve_problem(problem, initvals, verbose, **kwargs)
    return solution.x if solution.found else None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Test the "quadprog" QP solver on a model predictive control problem.

The problem is to balance a humanoid robot walking on a flat horizontal floor.
See the following post for context:

    https://scaron.info/robot-locomotion/prototyping-a-walking-pattern-generator.html
"""

import random
from dataclasses import dataclass
from typing import Optional

import numpy as np
import matplotlib.pyplot as plt
from scipy.sparse import csc_matrix

import qpsolvers
from qpsolvers import solve_qp

gravity = 9.81  # [m] / [s]^2


@dataclass
class HumanoidSteppingProblem:

    com_height: float = 0.8
    dsp_duration: float = 0.1
    end_pos: float = 0.3
    foot_length: float = 0.1
    horizon_duration: float = 2.5
    nb_timesteps: int = 16
    ssp_duration: float = 0.7
    start_pos: float = 0.0


class LinearModelPredictiveControl:

    """
    Linear model predictive control for a system with linear dynamics and
    linear constraints. This class is fully documented at:

        https://scaron.info/doc/pymanoid/walking-pattern-generation.html#pymanoid.mpc.LinearPredictiveControl
    """

    def __init__(
        self,
        A,
        B,
        C,
        D,
        e,
        x_init,
        x_goal,
        nb_timesteps: int,
        wxt: Optional[float],
        wxc: Optional[float],
        wu: float,
    ):
        assert C is not None or D is not None, "use LQR for unconstrained case"
        assert (
            wu > 0.0
        ), "non-negative control weight needed for regularization"
        assert wxt is not None or wxc is not None, "set either wxt or wxc"
        u_dim = B.shape[1]
        x_dim = A.shape[1]
        self.A = A
        self.B = B
        self.C = C
        self.D = D
        self.G = None
        self.P = None
        self.U = None
        self.U_dim = u_dim * nb_timesteps
        self.e = e
        self.h = None
        self.nb_timesteps = nb_timesteps
        self.q = None
        self.u_dim = u_dim
        self.wu = wu
        self.wxc = wxc
        self.wxt = wxt
        self.x_dim = x_dim
        self.x_goal = x_goal
        self.x_init = x_init
        #
        self.build()

    def build(self):
        phi = np.eye(self.x_dim)
        psi = np.zeros((self.x_dim, self.U_dim))
        G_list, h_list = [], []
        phi_list, psi_list = [], []
        for k in range(self.nb_timesteps):
            # Loop invariant: x == psi * U + phi * x_init
            if self.wxc is not None:
                phi_list.append(phi)
                psi_list.append(psi)
            C = self.C[k] if type(self.C) is list else self.C
            D = self.D[k] if type(self.D) is list else self.D
            e = self.e[k] if type(self.e) is list else self.e
            G = np.zeros((e.shape[0], self.U_dim))
            h = e if C is None else e - np.dot(C.dot(phi), self.x_init)
            if D is not None:
                # we rely on G == 0 to avoid a slower +=
                G[:, k * self.u_dim : (k + 1) * self.u_dim] = D
            if C is not None:
                G += C.dot(psi)
            if k == 0 and D is None:  # corner case, input has no effect
                assert np.all(h >= 0.0)
            else:  # regular case
                G_list.append(G)
                h_list.append(h)
            phi = self.A.dot(phi)
            psi = self.A.dot(psi)
            psi[:, self.u_dim * k : self.u_dim * (k + 1)] = self.B
        P = self.wu * np.eye(self.U_dim)
        q = np.zeros(self.U_dim)
        if self.wxt is not None and self.wxt > 1e-10:
            c = np.dot(phi, self.x_init) - self.x_goal
            P += self.wxt * np.dot(psi.T, psi)
            q += self.wxt * np.dot(c.T, psi)
        if self.wxc is not None and self.wxc > 1e-10:
            Phi = np.vstack(phi_list)
            Psi = np.vstack(psi_list)
            X_goal = np.hstack([self.x_goal] * self.nb_timesteps)
            c = np.dot(Phi, self.x_init) - X_goal
            P += self.wxc * np.dot(Psi.T, Psi)
            q += self.wxc * np.dot(c.T, Psi)
        self.P = P
        self.q = q
        self.G = np.vstack(G_list)
        self.h = np.hstack(h_list)
        self.P_csc = csc_matrix(self.P)
        self.G_csc = csc_matrix(self.G)

    def solve(self, solver: str, sparse: bool = False, **kwargs):
        P = self.P_csc if sparse else self.P
        G = self.G_csc if sparse else self.G
        U = solve_qp(P, self.q, G, self.h, solver=solver, **kwargs)
        self.U = U.reshape((self.nb_timesteps, self.u_dim))

    @property
    def states(self):
        assert self.U is not None, "you need to solve() the MPC problem first"
        X = np.zeros((self.nb_timesteps + 1, self.x_dim))
        X[0] = self.x_init
        for k in range(self.nb_timesteps):
            X[k + 1] = self.A.dot(X[k]) + self.B.dot(self.U[k])
        return X


class HumanoidModelPredictiveControl(LinearModelPredictiveControl):
    def __init__(self, problem: HumanoidSteppingProblem):
        T = problem.horizon_duration / problem.nb_timesteps
        nb_init_dsp_steps = int(round(problem.dsp_duration / T))
        nb_init_ssp_steps = int(round(problem.ssp_duration / T))
        nb_dsp_steps = int(round(problem.dsp_duration / T))
        state_matrix = np.array(
            [[1.0, T, T ** 2 / 2.0], [0.0, 1.0, T], [0.0, 0.0, 1.0]]
        )
        control_matrix = np.array([T ** 3 / 6.0, T ** 2 / 2.0, T])
        control_matrix = control_matrix.reshape((3, 1))
        zmp_from_state = np.array([1.0, 0.0, -problem.com_height / gravity])
        ineq_matrix = np.array([+zmp_from_state, -zmp_from_state])
        cur_max = problem.start_pos + 0.5 * problem.foot_length
        cur_min = problem.start_pos - 0.5 * problem.foot_length
        next_max = problem.end_pos + 0.5 * problem.foot_length
        next_min = problem.end_pos - 0.5 * problem.foot_length
        ineq_vector = [
            np.array([+1000.0, +1000.0])
            if i < nb_init_dsp_steps
            else np.array([+cur_max, -cur_min])
            if i - nb_init_dsp_steps <= nb_init_ssp_steps
            else np.array([+1000.0, +1000.0])
            if i - nb_init_dsp_steps - nb_init_ssp_steps < nb_dsp_steps
            else np.array([+next_max, -next_min])
            for i in range(problem.nb_timesteps)
        ]
        super().__init__(
            state_matrix,
            control_matrix,
            ineq_matrix,
            None,
            ineq_vector,
            x_init=np.array([problem.start_pos, 0.0, 0.0]),
            x_goal=np.array([problem.end_pos, 0.0, 0.0]),
            nb_timesteps=problem.nb_timesteps,
            wxt=1.0,
            wxc=None,
            wu=1e-3,
        )


def plot_mpc_solution(problem, mpc):
    t = np.linspace(0.0, problem.horizon_duration, problem.nb_timesteps + 1)
    X = mpc.states
    zmp_from_state = np.array([1.0, 0.0, -problem.com_height / gravity])
    zmp = X.dot(zmp_from_state)
    pos = X[:, 0]
    zmp_min = [x[0] if abs(x[0]) < 10 else None for x in mpc.e]
    zmp_max = [-x[1] if abs(x[1]) < 10 else None for x in mpc.e]
    zmp_min.append(zmp_min[-1])
    zmp_max.append(zmp_max[-1])
    plt.ion()
    plt.clf()
    plt.plot(t, pos)
    plt.plot(t, zmp, "r-")
    plt.plot(t, zmp_min, "g:")
    plt.plot(t, zmp_max, "b:")
    plt.grid(True)
    plt.show(block=True)


if __name__ == "__main__":
    problem = HumanoidSteppingProblem()
    mpc = HumanoidModelPredictiveControl(problem)
    mpc.solve(solver=random.choice(qpsolvers.available_solvers))
    plot_mpc_solution(problem, mpc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Apply lasso regularization to a quadratic program.

Details:
    https://scaron.info/blog/lasso-regularization-in-quadratic-programming.html
"""

import numpy as np

from qpsolvers import solve_qp

# Objective: || R x - s ||^2
n = 6
R = np.diag(range(1, n + 1))
s = np.ones(n)

# Convert our least-squares objective to quadratic programming
P = np.dot(R.transpose(), R)
q = -np.dot(s.transpose(), R)

# Linear inequality constraints: G x <= h
G = np.array(
    [
        [1.0, 0.0] * (n // 2),
        [0.0, 1.0] * (n // 2),
    ]
)
h = np.array([10.0, -10.0])

# Lasso parameter
t: float = 10.0

# Lasso: inequality constraints
G_lasso = np.vstack(
    [
        np.hstack([G, np.zeros((G.shape[0], n))]),
        np.hstack([+np.eye(n), -np.eye(n)]),
        np.hstack([-np.eye(n), -np.eye(n)]),
        np.hstack([np.zeros((1, n)), np.ones((1, n))]),
    ]
)
h_lasso = np.hstack([h, np.zeros(n), np.zeros(n), t])

# Lasso: objective
P_lasso = np.vstack(
    [
        np.hstack([P, np.zeros((n, n))]),
        np.zeros((n, 2 * n)),
    ]
)
q_lasso = np.hstack([q, np.ones(n)])

if __name__ == "__main__":
    x_unreg = solve_qp(P, q, G, h, solver="proxqp")
    print(f"Solution without lasso: {x_unreg = }")

    lasso_res = solve_qp(P_lasso, q_lasso, G_lasso, h_lasso, solver="proxqp")
    x_lasso = lasso_res[:n]
    z_lasso = lasso_res[n:]
    print(f"Solution with lasso ({t=}): {x_lasso = }")
    print(f"We can check that abs(x_lasso) = {z_lasso = }")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for SCS."""

import unittest
import warnings

import numpy as np

from qpsolvers import ProblemError

from .problems import get_sd3310_problem

try:
    from qpsolvers.solvers.scs_ import scs_solve_qp

    class TestSCS(unittest.TestCase):
        """Tests specific to SCS."""

        def test_problem(self):
            problem = get_sd3310_problem()
            P, q, G, h, A, b, lb, ub = problem.unpack()
            self.assertIsNotNone(scs_solve_qp(P, q, G, h, A, b, lb, ub))

        def test_unbounded_below(self):
            problem = get_sd3310_problem()
            P, q, _, _, _, _, _, _ = problem.unpack()
            P -= np.eye(3)  # make problem unbounded
            with self.assertRaises(ProblemError):
                scs_solve_qp(P, q)

except ImportError as exn:  # solver not installed
    warnings.warn(f"Skipping SCS tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for PIQP."""

import unittest
import warnings

from qpsolvers.exceptions import ParamError, ProblemError

from .problems import get_sd3310_problem

try:
    from qpsolvers.solvers.piqp_ import piqp_solve_qp

    class TestPIQP(unittest.TestCase):
        """Test fixture specific to the PIQP solver."""

        def test_dense_backend(self):
            """Try the dense backend."""
            problem = get_sd3310_problem()
            sol = piqp_solve_qp(
                problem.P,
                problem.q,
                problem.G,
                problem.h,
                problem.A,
                problem.b,
                backend="dense",
            )
            self.assertIsNotNone(sol)

        def test_sparse_backend(self):
            """Try the sparse backend."""
            problem = get_sd3310_problem()
            sol = piqp_solve_qp(
                problem.P,
                problem.q,
                problem.G,
                problem.h,
                problem.A,
                problem.b,
                backend="sparse",
            )
            self.assertIsNotNone(sol)

        def test_invalid_backend(self):
            """Exception raised when asking for an invalid backend."""
            problem = get_sd3310_problem()
            with self.assertRaises(ParamError):
                piqp_solve_qp(
                    problem.P,
                    problem.q,
                    problem.G,
                    problem.h,
                    problem.A,
                    problem.b,
                    backend="invalid",
                )

        def test_invalid_problems(self):
            """Exception raised when asking for an invalid backend."""
            problem = get_sd3310_problem()
            with self.assertRaises(ProblemError):
                piqp_solve_qp(
                    problem.P,
                    problem.q,
                    None,
                    problem.h,
                    problem.A,
                    problem.b,
                    backend="sparse",
                )
            with self.assertRaises(ProblemError):
                piqp_solve_qp(
                    problem.P,
                    problem.q,
                    problem.G,
                    None,
                    problem.A,
                    problem.b,
                    backend="sparse",
                )
            with self.assertRaises(ProblemError):
                piqp_solve_qp(
                    problem.P,
                    problem.q,
                    problem.G,
                    problem.h,
                    None,
                    problem.b,
                    backend="sparse",
                )
            with self.assertRaises(ProblemError):
                piqp_solve_qp(
                    problem.P,
                    problem.q,
                    problem.G,
                    problem.h,
                    problem.A,
                    None,
                    backend="sparse",
                )

except ImportError as exn:  # solver not installed
    warnings.warn(f"Skipping PIQP tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for ProxQP."""

import unittest
import warnings

from qpsolvers.exceptions import ParamError, ProblemError

from .problems import get_sd3310_problem

try:
    from qpsolvers.solvers.proxqp_ import proxqp_solve_qp

    class TestProxQP(unittest.TestCase):
        """Test fixture specific to the ProxQP solver."""

        def test_dense_backend(self):
            """Try the dense backend."""
            problem = get_sd3310_problem()
            proxqp_solve_qp(
                problem.P,
                problem.q,
                problem.G,
                problem.h,
                problem.A,
                problem.b,
                backend="dense",
            )

        def test_sparse_backend(self):
            """Try the sparse backend."""
            problem = get_sd3310_problem()
            proxqp_solve_qp(
                problem.P,
                problem.q,
                problem.G,
                problem.h,
                problem.A,
                problem.b,
                backend="sparse",
            )

        def test_invalid_backend(self):
            """Exception raised when asking for an invalid backend."""
            problem = get_sd3310_problem()
            with self.assertRaises(ParamError):
                proxqp_solve_qp(
                    problem.P,
                    problem.q,
                    problem.G,
                    problem.h,
                    problem.A,
                    problem.b,
                    backend="invalid",
                )

        def test_double_warm_start(self):
            """Exception when two warm-start values are provided."""
            problem = get_sd3310_problem()
            with self.assertRaises(ParamError):
                proxqp_solve_qp(
                    problem.P,
                    problem.q,
                    problem.G,
                    problem.h,
                    problem.A,
                    problem.b,
                    initvals=problem.q,
                    x=problem.q,
                )

        def test_invalid_inequalities(self):
            """Check for inconsistent parameters.

            Raise an exception in an implementation-dependent inconsistent set
            of parameters. This may happen when :func:`proxqp_solve_qp` it is
            called directly.
            """
            problem = get_sd3310_problem()
            with self.assertRaises(ProblemError):
                proxqp_solve_qp(
                    problem.P,
                    problem.q,
                    G=problem.G,
                    h=None,
                    lb=problem.q,
                )

except ImportError as exn:  # solver not installed
    warnings.warn(f"Skipping ProxQP tests: {exn}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Test an available QP solvers on a small problem with box inequalities."""

import random
from time import perf_counter

import numpy as np

from qpsolvers import available_solvers, print_matrix_vector, solve_qp

M = np.array([[1.0, 2.0, 0.0], [-8.0, 3.0, 2.0], [0.0, 1.0, 1.0]])
P = np.dot(M.T, M)  # this is a positive definite matrix
q = np.dot(np.array([3.0, 2.0, 3.0]), M)
A = np.array([1.0, 1.0, 1.0])
b = np.array([1.0])
lb = -0.5 * np.ones(3)
ub = 1.0 * np.ones(3)

x_sol = np.array([0.41463414566726164, -0.41463414566726164, 1.0])

if __name__ == "__main__":
    start_time = perf_counter()
    solver = random.choice(available_solvers)
    x = solve_qp(P, q, A=A, b=b, lb=lb, ub=ub, solver=solver)
    end_time = perf_counter()

    print("")
    print("    min. 1/2 x^T P x + q^T x")
    print("    s.t.   A * x == b")
    print("         lb <= x <= ub")
    print("")
    print_matrix_vector(P, "P", q, "q")
    print("")
    print_matrix_vector(A, "A", b, "b")
    print("")
    print_matrix_vector(lb.reshape((3, 1)), "lb", ub, "ub")
    print("")
    print(f"Solution: x = {x}")
    print(f"It should be close to x* = {x_sol}")
    print(f"Found in {1e6 * (end_time - start_time):.0f} [us] with {solver}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Solve linear least squares."""

from typing import Optional, Union

import numpy as np
import scipy.sparse as spa

from .problem import Problem
from .solve_qp import solve_qp


def __solve_dense_ls(
    R: Union[np.ndarray, spa.csc_matrix],
    s: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    W: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    solver: Optional[str] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    WR: Union[np.ndarray, spa.csc_matrix] = R if W is None else W @ R
    P = R.T @ WR
    q = -(s.T @ WR)
    if not isinstance(P, np.ndarray):
        P = P.tocsc()
    return solve_qp(
        P,
        q,
        G,
        h,
        A,
        b,
        lb,
        ub,
        solver=solver,
        initvals=initvals,
        verbose=verbose,
        **kwargs,
    )


def __solve_sparse_ls(
    R: Union[np.ndarray, spa.csc_matrix],
    s: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    W: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    solver: Optional[str] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    m, n = R.shape
    eye_m = spa.eye(m, format="csc")
    P = spa.block_diag(
        [spa.csc_matrix((n, n)), eye_m if W is None else W], format="csc"
    )
    q = np.zeros(n + m)
    P, q, G, h, A, b, lb, ub = Problem(P, q, G, h, A, b, lb, ub).unpack()
    if G is not None:
        G = spa.hstack([G, spa.csc_matrix((G.shape[0], m))], format="csc")
    if A is not None:
        A = spa.hstack([A, spa.csc_matrix((A.shape[0], m))], format="csc")
    Rx_minus_y = spa.hstack([R, -eye_m], format="csc")
    if A is not None and b is not None:  # help mypy
        A = spa.vstack([A, Rx_minus_y], format="csc")
        b = np.hstack([b, s])
    else:  # no input equality constraint
        A = Rx_minus_y
        b = s
    if lb is not None:
        lb = np.hstack([lb, np.full((m,), -np.inf)])
    if ub is not None:
        ub = np.hstack([ub, np.full((m,), np.inf)])
    xy = solve_qp(
        P,
        q,
        G,
        h,
        A,
        b,
        lb,
        ub,
        solver=solver,
        initvals=initvals,
        verbose=verbose,
        **kwargs,
    )
    return xy[:n] if xy is not None else None


def solve_ls(
    R: Union[np.ndarray, spa.csc_matrix],
    s: np.ndarray,
    G: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    W: Optional[Union[np.ndarray, spa.csc_matrix]] = None,
    solver: Optional[str] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    sparse_conversion: Optional[bool] = None,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a constrained weighted linear Least Squares problem.

    The linear least squares is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{x}{\mbox{minimize}} &
                \frac12 \| R x - s \|^2_W
                = \frac12 (R x - s)^T W (R x - s) \\
            \mbox{subject to}
                & G x \leq h          \\
                & A x = b             \\
                & lb \leq x \leq ub
        \end{array}\end{split}

    using the QP solver selected by the ``solver`` keyword argument.

    Parameters
    ----------
    R :
        Union[np.ndarray, spa.csc_matrix] factor of the cost function (most
        solvers require :math:`R^T W R` to be positive definite, which means
        :math:`R` should have full row rank).
    s :
        Vector term of the cost function.
    G :
        Linear inequality matrix.
    h :
        Linear inequality vector.
    A :
        Linear equality matrix.
    b :
        Linear equality vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    W :
        Definite symmetric weight matrix used to define the norm of the cost
        function. The standard L2 norm (W = Identity) is used by default.
    solver :
        Name of the QP solver, to choose in
        :data:`qpsolvers.available_solvers`. This argument is mandatory.
    initvals :
        Vector of initial `x` values used to warm-start the solver.
    verbose :
        Set to `True` to print out extra information.
    sparse_conversion :
        Set to `True` to use a sparse conversion strategy and to `False` to use
        a dense strategy. By default, the conversion strategy to follow is
        determined by the sparsity of :math:`R` (sparse if CSC matrix, dense
        otherwise). See Notes below.

    Returns
    -------
    :
        Optimal solution if found, otherwise ``None``.

    Note
    ----
    Some solvers (like quadprog) will require a full-rank matrix :math:`R`,
    while others (like ProxQP or QPALM) can work even when :math:`R` has a
    non-empty nullspace.

    Notes
    -----
    This function implements two strategies to convert the least-squares cost
    :math:`(R, s)` to a quadratic-programming cost :math:`(P, q)`: one that
    assumes :math:`R` is dense, and one that assumes :math:`R` is sparse. These
    two strategies are detailed in `this note
    <https://scaron.info/blog/conversion-from-least-squares-to-quadratic-programming.html>`__.
    The sparse strategy introduces extra variables :math;`y = R x` and will
    likely perform better on sparse problems, although this may not always be
    the case (for instance, it may perform worse if :math:`R` has many more
    rows than columns).

    Extra keyword arguments given to this function are forwarded to the
    underlying solvers. For example, OSQP has a setting `eps_abs` which we can
    provide by ``solve_ls(R, s, G, h, solver='osqp', eps_abs=1e-4)``.
    """
    if sparse_conversion is None:
        sparse_conversion = not isinstance(R, np.ndarray)
    if sparse_conversion:
        return __solve_sparse_ls(
            R, s, G, h, A, b, lb, ub, W, solver, initvals, verbose, **kwargs
        )
    return __solve_dense_ls(
        R, s, G, h, A, b, lb, ub, W, solver, initvals, verbose, **kwargs
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Make sure problem matrices are sparse."""

import warnings
from typing import Optional, Tuple, Union

import numpy as np
import scipy.sparse as spa

from ..warnings import SparseConversionWarning


def __warn_about_sparse_conversion(matrix_name: str, solver_name: str) -> None:
    """Warn about conversion from dense to sparse matrix.

    Parameters
    ----------
    matrix_name :
        Name of matrix being converted from dense to sparse.
    solver_name :
        Name of the QP solver matrices will be passed to.
    """
    warnings.warn(
        f"Converted matrix '{matrix_name}' of your problem to "
        f"scipy.sparse.csc_matrix to pass it to solver '{solver_name}'; "
        f"for best performance, build your matrix as a csc_matrix directly.",
        category=SparseConversionWarning,
    )


def ensure_sparse_matrices(
    solver_name: str,
    P: Union[np.ndarray, spa.csc_matrix],
    G: Optional[Union[np.ndarray, spa.csc_matrix]],
    A: Optional[Union[np.ndarray, spa.csc_matrix]],
) -> Tuple[spa.csc_matrix, Optional[spa.csc_matrix], Optional[spa.csc_matrix]]:
    """
    Make sure problem matrices are sparse.

    Parameters
    ----------
    solver_name :
        Name of the QP solver matrices will be passed to.
    P :
        Cost matrix.
    G :
        Inequality constraint matrix, if any.
    A :
        Equality constraint matrix, if any.

    Returns
    -------
    :
        Tuple of all three matrices as sparse matrices.
    """
    if isinstance(P, np.ndarray):
        __warn_about_sparse_conversion("P", solver_name)
        P = spa.csc_matrix(P)
    if isinstance(G, np.ndarray):
        __warn_about_sparse_conversion("G", solver_name)
        G = spa.csc_matrix(G)
    if isinstance(A, np.ndarray):
        __warn_about_sparse_conversion("A", solver_name)
        A = spa.csc_matrix(A)
    return P, G, A



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2016-2022 Stéphane Caron and the qpsolvers contributors

"""Unit tests for KVXOPT."""

import unittest
import warnings
from typing import Tuple

import numpy as np
import scipy.sparse as spa
from numpy import array, ones
from numpy.linalg import norm
from qpsolvers.problems import get_qpsut01

from .problems import get_sd3310_problem

try:
    import kvxopt
    from qpsolvers.solvers.kvxopt_ import kvxopt_solve_problem, kvxopt_solve_qp

    class TestKVXOPT(unittest.TestCase):
        """Test fixture for the KVXOPT solver."""

        def setUp(self):
            """Prepare test fixture."""
            warnings.simplefilter("ignore", category=UserWarning)

        def get_sparse_problem(
            self,
        ) -> Tuple[kvxopt.matrix, np.ndarray, kvxopt.matrix, np.ndarray]:
            """Get sparse problem as a quadruplet of values to unpack.

            Returns
            -------
            P :
                Symmetric cost matrix.
            q :
                Cost vector.
            G :
                Linear inequality matrix.
            h :
                Linear inequality vector.
            """
            n = 150
            M = spa.lil_matrix(spa.eye(n))
            for i in range(1, n - 1):
                M[i, i + 1] = -1
                M[i, i - 1] = 1
            P = spa.csc_matrix(M.dot(M.transpose()))
            q = -ones((n,))
            G = spa.csc_matrix(-spa.eye(n))
            h = -2.0 * ones((n,))
            return P, q, G, h

        def test_sparse(self):
            """Test KVXOPT on a sparse problem."""
            P, q, G, h = self.get_sparse_problem()
            x = kvxopt_solve_qp(P, q, G, h)
            self.assertIsNotNone(x, 'solver="kvxopt"')
            known_solution = array([2.0] * 149 + [3.0])
            sol_tolerance = 1e-2  # aouch, not great!
            self.assertLess(
                norm(x - known_solution), sol_tolerance, 'solver="kvxopt"'
            )
            self.assertLess(max(G.dot(x) - h), 1e-10, 'solver="kvxopt"')

        def test_extra_kwargs(self):
            """Call KVXOPT with various solver-specific settings."""
            problem = get_sd3310_problem()
            x = kvxopt_solve_qp(
                problem.P,
                problem.q,
                problem.G,
                problem.h,
                problem.A,
                problem.b,
                maxiters=10,
                abstol=1e-1,
                reltol=1e-1,
                feastol=1e-2,
                refinement=3,
            )
            self.assertIsNotNone(x, 'solver="kvxopt"')

        def test_infinite_linear_bounds(self):
            """KVXOPT does not yield a domain error on infinite bounds."""
            problem, _ = get_qpsut01()
            problem.h[1] = +np.inf
            x = kvxopt_solve_problem(problem)
            self.assertIsNotNone(x, 'solver="kvxopt"')

        def test_infinite_box_bounds(self):
            """KVXOPT does not yield a domain error infinite box bounds."""
            problem, _ = get_qpsut01()
            problem.lb[1] = -np.inf
            problem.ub[1] = +np.inf
            x = kvxopt_solve_problem(problem)
            self.assertIsNotNone(x, 'solver="kvxopt"')

except ImportError as exn:  # in case the solver is not installed
    warnings.warn(f"Skipping KVXOPT tests: {exn}")


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright 2024 Lev Kozlov
"""
Solver interface for `qpax <https://github.com/kevin-tracy/qpax>`__.

qpax is an open-source QP solver that can be combined with JAX's jit and vmap
functionality, as well as differentiated with reverse-mode differentiation. It
is based on a primal-dual interior point algorithm. If you are using qpax in
a scientific work, consider citing the corresponding paper [Tracy2024]_.
"""

import warnings
from typing import Optional

import numpy as np
import qpax
import scipy.sparse as spa

from ..conversions import linear_from_box_inequalities, split_dual_linear_box
from ..exceptions import ProblemError
from ..problem import Problem
from ..solution import Solution


def qpax_solve_problem(
    problem: Problem,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Solution:
    """Solve a quadratic program using qpax.

    Parameters
    ----------
    problem :
        Quadratic program to solve.
    initvals :
        Warm-start guess vector (not used).
    verbose :
        Set to `True` to print out extra information.

    Returns
    -------
    :
        Solution to the QP returned by the solver.

    Notes
    -----
    All other keyword arguments are forwarded as options to qpax. For
    instance, you can call ``qpax_solve_qp(P, q, G, h, solver_tol=1e-5)``.
    For a quick overview, the solver accepts the following settings:

    .. list-table::
       :widths: 30 70
       :header-rows: 1

       * - Name
         - Effect
       * - ``solver_tol``
         - Tolerance for the solver.

    Note that `jax` by default uses 32-bit floating point numbers, which can
    lead to numerical instability. If you encounter numerical issues, consider
    using 64-bit floating point numbers by setting
    ```python
    import jax
    jax.config.update("jax_enable_x64", True)
    ```
    """
    P, q, G, h, A, b, lb, ub = problem.unpack()
    n: int = q.shape[0]

    if initvals is not None and verbose:
        warnings.warn("warm-start values are ignored by qpax")

    if G is None and h is not None:
        raise ProblemError(
            "Inconsistent inequalities: G is not set but h is set"
        )
    if G is not None and h is None:
        raise ProblemError("Inconsistent inequalities: G is set but h is None")
    if A is None and b is not None:
        raise ProblemError(
            "Inconsistent inequalities: A is not set but b is set"
        )
    if A is not None and b is None:
        raise ProblemError("Inconsistent inequalities: A is set but b is None")

    # construct the qpax problem
    G, h = linear_from_box_inequalities(G, h, lb, ub, use_sparse=False)
    if G is None:
        G = np.zeros((0, n))
        h = np.zeros((0,))

    # qpax does not support A, b to be None.
    A_qpax = np.zeros((0, n)) if A is None else A
    b_qpax = np.zeros((0)) if b is None else b

    # qpax does not support sparse matrices,
    # so we need to convert them to dense
    if isinstance(P, spa.csc_matrix):
        P = P.toarray()
    if isinstance(A_qpax, spa.csc_matrix):
        A_qpax = A_qpax.toarray()
    if isinstance(G, spa.csc_matrix):
        G = G.toarray()

    x, s, z, y, converged, iters1 = qpax.solve_qp(
        P,
        q,
        A_qpax,
        b_qpax,
        G,
        h,
        **kwargs,
    )

    solution = Solution(problem)
    solution.x = x
    solution.found = converged
    solution.y = y

    # split the dual variables into
    # the box constraints and the linear constraints
    solution.z, solution.z_box = split_dual_linear_box(
        z, problem.lb, problem.ub
    )

    # store information about the solution
    # and the resulted raw variables in the extras
    solution.extras = {
        "info": {
            "iterations": iters1,
            "converged": converged,
        },
        "variables": {
            "x": x,
            "y": y,
            "z": z,
            "s": s,
        },
    }

    return solution


def qpax_solve_qp(
    P: np.ndarray,
    q: np.ndarray,
    G: Optional[np.ndarray] = None,
    h: Optional[np.ndarray] = None,
    A: Optional[np.ndarray] = None,
    b: Optional[np.ndarray] = None,
    lb: Optional[np.ndarray] = None,
    ub: Optional[np.ndarray] = None,
    initvals: Optional[np.ndarray] = None,
    verbose: bool = False,
    **kwargs,
) -> Optional[np.ndarray]:
    r"""Solve a quadratic program using qpax.

    The quadratic program is defined as:

    .. math::

        \begin{split}\begin{array}{ll}
        \underset{\mbox{minimize}}{x} &
            \frac{1}{2} x^T P x + q^T x \\
        \mbox{subject to}
            & G x \leq h                \\
            & A x = b                   \\
            & lb \leq x \leq ub
        \end{array}\end{split}

    It is solved using `qpax
    <https://github.com/kevin-tracy/qpax>`__.
    `Paper: <https://arxiv.org/pdf/2406.11749>`__.

    Parameters
    ----------
    P :
        Positive semidefinite cost matrix.
    q :
        Cost vector.
    G :
        Linear inequality constraint matrix.
    h :
        Linear inequality constraint vector.
    A :
        Linear equality constraint matrix.
    b :
        Linear equality constraint vector.
    lb :
        Lower bound constraint vector.
    ub :
        Upper bound constraint vector.
    verbose :
        Set to `True` to print out extra information.
    initvals :
        Warm-start guess vector. Not used.

    Returns
    -------
    :
        Primal solution to the QP, if found, otherwise ``None``.
    """
    problem = Problem(P, q, G, h, A, b, lb, ub)
    solution = qpax_solve_problem(problem, initvals, verbose, **kwargs)
    return solution.x if solution.found else None
